using DataFrames
using XLSX
using JuMP  # Julia for Mathematical Programming
using Gadfly
using Statistics

# Data Import
file = ("data_final.xlsx")

contiguity = DataFrame(XLSX.readtable(file, "contiguity")...)
area_bu =DataFrame(XLSX.readtable(file, "area")...)
distance_bu = DataFrame(XLSX.readtable(file, "network_distance")...)
city_subunit1 =  DataFrame(XLSX.readtable(file, "city")...)
care_load=DataFrame(XLSX.readtable(file, "workload")...)
compa = DataFrame(XLSX.readtable("agDATA.xlsx", "compatibility")...)
e2 = compa[1:84, 2:85]
#Parameters
N = 84
M = 4 # enter the number of districts to design here
s = 6/5 # enter the average speed here
e = contiguity[1:84, 2:85]
area=area_bu[1:84,2]
d=distance_bu[1:84,2:85] #d[i,j] be the network distance between subunit i and subunit j.
city_subunit = city_subunit1[1:84, 2]
number =care_load[1:84,2] #Number of patient
wt =care_load[1:84,3] # Care load  OF SUBREGION i in minutes


#Travel Load Estimator
function travelload(N,A)
    # Calculate travel load based on number of people and area
    return (0.75 * sqrt(N*A) *s)
end
######################################################################################
##################LOWER BOUND ##################
######################################################################################
#Most high load

most_loaded_sub=maximum(wt)
least_loaded_sub=minimum(wt)
total_number_patient=sum(number)
total_area=sum(area)
#If M=4
patient_points_per_district=total_number_patient/M
mean_area_per_district=total_area/M
#Travel_load
travel_lb=travelload(patient_points_per_district,mean_area_per_district)
mean_careload=sum(wt)/M
total_lb_load=travel_lb+mean_careload
if most_loaded_sub>total_lb_load
    lb=most_loaded_sub-total_lb_load
else
    lb=0
end


######################################################################################
##################PHASE 1 : PAP => Incompability   ##################
######################################################################################
#1)We define Incompability
time1initial=time()

w1=0.1 #Distance weight
w2=2#Contiguity weight
w3=1#City weight

function S(i,j)
    if city_subunit[i]==city_subunit[j]
        city=1
    else
        city=0
    end
    return (w2*(1-e[i,j]))+(w3*(1-city))+(w1*d[i,j])
end

matrixIncompa = Vector{Float64}()

for i in 1:N
    ind = sum(j ≠ i ? S(i, j) : 0 for j in 1:N)
    push!(matrixIncompa, ind)
end
 #For each basic unit we compute the total number of cities with wich the given city is contiguous or in the same city

MatrixRelation=zeros(N, N)

for i in 1:N
    for j in 1:N
        if i != j
            MatrixRelation[i, j] = S(i, j)
        end
    end
end

limit=4 #The limit #Means that if not contiguous and not same city , it is in max 10km

MatrixI=zeros(N, N) #If 1 it means that incompatible
for i in 1:N
    for j in 1:N
        if MatrixRelation[i, j]>4
            MatrixI[i, j] = 1
        end
    end
end

scoreIncompa = Vector{Float64}()
for i in 1:N
    ind = sum(j ≠ i ? MatrixI[i,j] : 0 for j in 1:N)
    push!(scoreIncompa, ind)
end

matrices = Vector{Vector{Int}}(undef, M) #We create a matrix with M vectors in wich each vecotr represent a district
for i in 1:M
    matrices[i] = []
end

sorted_incompa = sortperm(scoreIncompa, rev=true)  # The first = most incompatible
i = 1
t = 1
while i <= M
    sub1 = sorted_incompa[t]  # Index of the max incompatibility
    if i == 1  # We select the most incompatible and place it in the first district / first element
        push!(matrices[i], sub1)
        t += 1
        i += 1
    else  # For the other ones, we make sure there is no contiguity with previously selected basic units
        noncontig = 0
        x = 1
        a = i - 1
        while x <= a
            if d[matrices[x][1], sub1] >= 22
                noncontig += 1
                if noncontig == a && x == a  # We have verified for all previous basic units
                    push!(matrices[i], sub1)
                    t += 1
                    i += 1  # Move to the next matrix
                    break
                else
                    x += 1
                end
            else  # Case where the subunit is contiguous or in the same city as a previous one
                t += 1
                break  # Exit the inner while loop
            end
        end
    end
end
#=
#To check if each element is unique
function check_unique_elements(matrices)
    for i in eachindex(matrices)
        for j in eachindex(matrices[i])
            for k in (j+1):length(matrices[i])
                if matrices[i][j] == matrices[i][k]
                    return false
                end
            end
        end
    end

    return true
end


check_unique_elements(matrices)
=#


######################################################################################
###############PHASE 2: Workload balancing heuristic/Initial solution#################
######################################################################################
# Initialize an empty array to hold the workloads of each basic unit
workloadofbu = []

# Loop over each basic unit and compute the total workload
for i in 1:length(number)
    tot_load = travelload(number[i], area[i]) + wt[i]
    push!(workloadofbu, tot_load)
end


# Sort basic units in decreasing order of workloads
index_decreasing = sortperm(workloadofbu, rev=true) #Gives the load index in descending order

# Remove any existing basic units in districts from index_decreasing
for j in 1:M
    deleteat!(index_decreasing, findfirst(x -> x == matrices[j][1], index_decreasing))
end

#3) Select the first basic unit of the list and assign it to the least loaded district containing one basic unit contiguous with the first element of the list.
# Initialize workload of each district to 0
loadofdistrict = zeros(M)

# Compute the workload of each initial district
for i in 1:M
    loadofdistrict[i] = workloadofbu[matrices[i][1]]
end
# Loop until all basic units have been assigned to districts

a = 1
while !isempty(index_decreasing)
    i = 1
    assigned = false  # Initialize assigned flag

    while i <= M
        # Find the least loaded district
        sorted_district = sortperm(loadofdistrict, rev=false)
        least_index = sorted_district[i]

        # Check conditions for assignment
        found_assignment = false
        for j in 1:size(matrices[least_index])[1]
            if e[index_decreasing[a], matrices[least_index][j]]==1
                # Assign the first element of the list to the least loaded district
                adid=index_decreasing[a]
                append!(matrices[least_index],adid)
                deleteat!(index_decreasing, a)
                # Recompute the workload of the least loaded district
                sum_number = sum(number[matrices[least_index]])
                sum_area = sum(area[matrices[least_index]])
                cload = sum(wt[matrices[least_index]])
                loadofdistrict[least_index] = travelload(sum_area, sum_number) + cload
                assigned = true  # Set assigned flag to true
                found_assignment = true
                break
            end
        end

        if found_assignment
            a = 1
            break
        else
            i += 1
        end
    end

    if !assigned
        a += 1
    end
end
time2initial=time()
timeinitial=time2initial-time1initial

Objectiveinitial = maximum(loadofdistrict)-minimum(loadofdistrict)
initial=copy(matrices)
allunique(matrices)
initial_load=copy(loadofdistrict)
#=
####Just careload
# Initialize an empty array to hold the workloads of each basic unit
careloadofdistrictInitial = []

# Loop over each basic unit and compute the total workload
for i in 1:M
    sum_c_i=0
    sum_c_i = sum(wt[matrices[i]])
    push!(careloadofdistrictInitial, sum_c_i)
end
objective_care_Initial=maximum(careloadofdistrictInitial)-minimum(careloadofdistrictInitial)
=#
######################################################################################
############Phase 3 : Sol Improvement Heuristic##############
######################################################################################

time1descent1=time()
 test = 0
 value1=[]
 mat_test = []
 mat_select = []
 initial_load_bu = copy(workloadofbu)
 iter = 0
 improved = 0
 xx = M
 f = 1
 pp = sortperm(loadofdistrict, rev=true)
 selected_last_min = 0
 selected_last_max =0
 limit=10000
 debut=0
 tracking=[]
 num=0
 while iter<limit
     xx = M
     f = 1
     pp = sortperm(loadofdistrict, rev=true)
     least_index = pp[xx] ##On revient ici
     most_index = pp[f]
     if length(matrices[most_index]) == 1
         for i in 1:M-2
             if length(matrices[most_index+i])>1
                 f=f+1
                 break
             end
         end
     end   #In order to avoid to have an infinite
     most_index = pp[f]
     most_value = loadofdistrict[most_index]
     least_value = loadofdistrict[least_index]
     difference = most_value - least_value
     aa = copy(matrices[most_index])
     idx = argmin(abs.(workloadofbu[aa] .- difference))
     debut=debut+1
     con=true
     while con==true
         last = length(mat_select)
         if last > 1
             if mat_select[last] == aa[idx] && length(aa) > 1  ||  mat_select[last-1] == aa[idx] && length(aa) > 1
                 deleteat!(aa, idx)
                 idx = argmin(abs.(workloadofbu[aa] .- difference))
             elseif length(aa) <= 1 && xx > 2
                 #On tourne en rond
                 #on doit transférer le deuxieme moins loadé #mais du coup celui la peut tourner en rond après
                 xx = xx - 1
                 least_index = pp[xx]
                 most_index = pp[f]
                 least_value = loadofdistrict[least_index]
                 most_value = loadofdistrict[most_index]
                 difference = most_value - least_value
                 aa = copy(matrices[most_index])
                 idx = argmin(abs.(workloadofbu[aa] .- difference))
             elseif xx==2 && length(aa) <= 1
                 if f<M
                     f=f+1
                     most_index = pp[f]
                     least_value=pp[xx]
                     most_value = loadofdistrict[most_index]
                     difference = most_value - least_value #Recalculer ou pas?
                     aa = copy(matrices[most_index])
                     idx = argmin(abs.(workloadofbu[aa] .- difference))
                 else
                 end
                 #We can't add aa, we can't xx either, so we'll go get another aa
                #we have to do something else
             end
         end

         selected_last_min = least_index
         selected_last_max = most_index
         println("first")
         println(idx)
         jcomp = 0
         size_matrices_selected = size(matrices[least_index], 1)

         for j in 1:size_matrices_selected
             if e[aa[idx], matrices[least_index][j]] == 0
                 jcomp += 1
                 break
             end
         end

         if jcomp > 0
             init_mostload=loadofdistrict[most_index]
             init_leastload=loadofdistrict[least_index]
             if length(value1)==0
                 difference_test = maximum(loadofdistrict)-minimum(loadofdistrict)
             else
                 difference_test=value1[length(value1)]
             end
             push!(matrices[least_index], aa[idx])
             sup = findfirst(matrices[most_index] .== aa[idx])
             deleteat!(matrices[most_index], sup)
             sum_number_least = sum(number[matrices[least_index]])
             sum_number_most = sum(number[matrices[most_index]])
             sum_area_least = sum(area[matrices[least_index]])
             sum_area_most = sum(area[matrices[most_index]])
             cload_least = sum(wt[matrices[least_index]])
             cload_most = sum(wt[matrices[most_index]])
             ll1= (travelload(sum_area_least, sum_number_least) + cload_least)
             ll2 = (travelload(sum_area_most, sum_number_most) + cload_most)
             loadofdistrict[least_index] = ll1
             loadofdistrict[most_index] = ll2
             objective=maximum(loadofdistrict)-minimum(loadofdistrict)
             #New value
             num=num+1
             append!(tracking,objective)
             if difference_test>=objective
                 # Accept the move
                 push!(value1,objective)
                 improved=improved+1
                 push!(mat_select, aa[idx])
                     println("v6")
                 xx=M
                 f=1
                 con=false
                 break
             else
                 # Revert the move
                 push!(matrices[most_index], aa[idx])
                 sup = findfirst(matrices[least_index] .== aa[idx])
                 deleteat!(matrices[least_index], sup)
                 loadofdistrict[least_index] =init_leastload
                 loadofdistrict[most_index]=init_mostload
                 push!(mat_select, aa[idx])
                 #On doit regarder si on peut changer un autre élément  #peut etre faire matrice
                 global iter=iter+1
                 if iter >= limit
                      # Exit the loop when iter reaches or exceeds 100
                 break
                 end
                 if length(aa) > 1 #Ici on tranfere entre mm matrice donc mm diff
                     deleteat!(aa, idx)
                     idx = argmin(abs.(workloadofbu[aa] .- difference))
                     last = length(mat_select)
                     break
                 end
             end
         elseif length(aa) > 1 #Ici on tranfere entre mm matrice donc mm diff
             deleteat!(aa, idx)
             idx = argmin(abs.(workloadofbu[aa] .- difference))
             last = length(mat_select)
             break

         elseif length(aa) <= 1
             if xx > 2 #least index change #diff change
                 #Same diff we will check the nex lest loaded
                 aa = copy(matrices[most_index])
                 pp = sortperm(loadofdistrict, rev=true)
                 xx -= 1
                 least_index = pp[xx]
                 difference = pp[f] - least_index
                 idx = argmin(abs.(workloadofbu[aa] .- difference))
             else
                 break
             end
         end
     end
 end


 time2descent1=time()
 timedescent=time2descent1-time1descent1
 ppv = sortperm(loadofdistrict, rev=true) #The first the most loaded matrices index
 least_index_cv = ppv[M]
 most_index_cv = ppv[1]
 most_value_cv = loadofdistrict[most_index_cv]
 least_value_cv = loadofdistrict[least_index_cv]
ObjectiveDescent = most_value_cv - least_value_cv
#=
####Just careload
# Initialize an empty array to hold the workloads of each basic unit
careloadofdistrictDescent = []

# Loop over each basic unit and compute the total workload
for i in 1:M
    sum_c_desc=0
    sum_c_desc = sum(wt[matrices[i]])
    push!(careloadofdistrictDescent, sum_c_desc)
end
objective_care_descent=maximum(careloadofdistrictDescent)-minimum(careloadofdistrictDescent)
=#

######################################################################################
########Phase 4 : Simulated Annealing #################
######################################################################################

time1 = time() #STARTING TIME
value2 = []
mat_select = []
initial_load_bu = copy(workloadofbu)
iter = 1
xx = M
pp = sortperm(loadofdistrict, rev=true)
selected_last_min = 0
selected_last_max = 0
tracking = [] # for graph
T=((ObjectiveDescent-lb)/M)
alpha = 0.9# Cooling factor
L = N*(M)# Length of a plateau
iteration = 0
plateau = 1
k=0
K2 = 5 # NUMBER F PLATEAUr
epsilon = 5
improv_within_plateau = false # At the beginning, we have no improvement within a plateau
iteration_within_plateau = 1 # Used to compute the number of iterations within a plateau (since the last decrease)
nb_plateau_without_improv = 0 # Number of plateau where no improvement has been made
plateau_nb = 1 # Will count the number of the plateau
acceptedmove_within_plateau = 0 # Variable computing the number of accepted moves within a plateau
acceptedmove_for_K2_plateau = Array{Float64}(undef, K2) # Vector where we store the number of accepted moves per plateau for the last K2 plateau
improv=false
energy=[]
improved=false
# Define the acceptance probability function
function acceptance_probability(delta, temperature)
    return exp(-delta / temperature)
end

Terminate=false
while Terminate==false
    k += 1
    xx=M
    f=1
    pp = sortperm(loadofdistrict, rev=true)
    least_index = pp[xx]
    most_index = pp[f]

    if length(matrices[most_index]) == 1
        f += 1
    end
    most_index = pp[f]
    global aa = copy(matrices[most_index])
    idxv = rand(aa)
    idx = findfirst(aa .== idxv)
    last = length(mat_select)
    println("v1")

    if last > 1
        if (mat_select[last] == idxv && length(aa) > 1) || (mat_select[last - 1] == idxv && length(aa) > 1)
            deleteat!(aa, idx)
            idxv = rand(aa)
            idx = findfirst(aa .== idxv)
            println("v2")
        elseif length(aa) <= 1 && selected_last_min == most_index && selected_last_max == least_index && xx > 2
            xx -= 1
            least_index = pp[xx]
            println("v3")
        end
    end
    con=true
    while con==true
    jcomp = 0
    size_matrices_selected = size(matrices[least_index], 1)
    for j in 1:size_matrices_selected
        if e[idxv, matrices[least_index][j]] == 1
            jcomp += 1
            break
        end
    end
    println("v4")

    if jcomp > 0
        init_mostload=loadofdistrict[most_index]
        init_leastload=loadofdistrict[least_index]

        if length(value2)==0
            difference_test = maximum(loadofdistrict)-minimum(loadofdistrict)
        else
            difference_test=value2[length(value2)]
        end
        push!(matrices[least_index], idxv)
        sup = findfirst(matrices[most_index] .== idxv)
        deleteat!(matrices[most_index], sup)
        println("second")
        sum_number_least = sum(number[matrices[least_index]])
        sum_number_most = sum(number[matrices[most_index]])
        sum_area_least = sum(area[matrices[least_index]])
        sum_area_most = sum(area[matrices[most_index]])
        cload_least = sum(wt[matrices[least_index]])
        cload_most = sum(wt[matrices[most_index]])
        ll1= (travelload(sum_area_least, sum_number_least) + cload_least)
        ll2 = (travelload(sum_area_most, sum_number_most) + cload_most)
        loadofdistrict[least_index] = ll1
        loadofdistrict[most_index] = ll2
        objective=maximum(loadofdistrict)-minimum(loadofdistrict)

        #Nouvelle valeur
        deltaF=difference_test-objective
        acceptance_prob = acceptance_probability(-deltaF, T)
        append!(energy, acceptance_prob)
        append!(tracking,objective)
        if difference_test>=objective
            push!(value2,objective)
            improved=improved+1
            push!(mat_select, idxv)
            acceptedmove_within_plateau=acceptedmove_within_plateau+1
            improv=true
            con=false
            break
        elseif rand() <= acceptance_prob
            push!(value2,objective)
            improved=improved+1
            push!(mat_select, idxv)
            acceptedmove_within_plateau=acceptedmove_within_plateau+1
            improv=false
            con=false
            break
        else
            improv=false
            push!(matrices[most_index],idxv )
            sup = findfirst(matrices[least_index] .==idxv )
            deleteat!(matrices[least_index], sup)
            loadofdistrict[least_index] =init_leastload
            loadofdistrict[most_index]=init_mostload
            push!(mat_select, idxv)
            con=false #Ca va retourner au début #Ca va prendre les meme choses !
            break

        end
        elseif length(aa) > 1 #Ici on tranfere entre mm matrice donc mm diff
            deleteat!(aa, idx)
            idxv=rand(aa)
            idx=findfirst(aa .== idxv)
            last = length(mat_select)

            if last > 1 #On s'assure de ne pas prendre le meme #mm diff
                if (mat_select[last-i] ==idxv  && length(aa) > 1 )  || (mat_select[last-1] ==idxv  && length(aa) > 1 ) # Le dernier qu'on avait sélectionné
                        deleteat!(aa, idx)
                        idxv=rand(aa)
                        idx=findfirst(aa .== idxv)
                else
                end
            else
            end

        elseif length(aa) <= 1
            if xx > 2 #least index change #diff change
                #la diff est mm , on va mettre vers la deuxieme moins loadé
                global aa = copy(matrices[most_index])
                pp = sortperm(loadofdistrict, rev=true)
                xx -= 1
                least_index = pp[xx]
                idxv=rand(aa)
                idx=findfirst(aa .== idxv)
            end
        end
    end

    if iteration_within_plateau < L
        iteration_within_plateau += 1
        if improv==true
            improv_within_plateau = true
        end
    else
        if improv==true
            improv_within_plateau = true
        end

        if improv_within_plateau == false
            nb_plateau_without_improv += 1
        else
            nb_plateau_without_improv=0
        end
        if plateau_nb <= K2 # If we haven't go through K2 plateau, we continue to fill in order the vector acceptedmove_for_K2_plateau
            acceptedmove_for_K2_plateau[plateau_nb] = acceptedmove_within_plateau # Contains all accepted moves of LAST K2 plateau
        else # If we have already go through K2 plateau
            # move all elements one place to the left in order to remove the element from the farthest plateau and replace by the number of accepted moves of the newest plateau
            for i in 1:K2
                if i != K2
                    acceptedmove_for_K2_plateau[i] = acceptedmove_for_K2_plateau[i + 1]
                else
                    acceptedmove_for_K2_plateau[i] = acceptedmove_within_plateau
                end
            end
        end

        nb_plateau_respect_condition=0
        for i in 1:K2
            if acceptedmove_for_K2_plateau[i]<((epsilon/100)*L)
                nb_plateau_respect_condition=nb_plateau_respect_condition+1
            end
        end
        # STOP DECISION

        if (nb_plateau_without_improv >= K2) && (nb_plateau_respect_condition==K2)
            Terminate=true

        else
            # Proceed to the next plateau
            iteration_within_plateau=1 # reinitialize the iteration count as we ended the plateau
            acceptedmove_within_plateau = 0 # reset the number of moves on the plateau as we ended it
            T=T*alpha # decrease the temperature by alpha

            #println("T=", T)
            improv_within_plateau=false # reinitialize the fact that we have made an improvement within the plateau
            plateau_nb=plateau_nb+1
        end
    end
end

time2 = time() # END TIME
FINALTIME=time2-time1
println("Execution time: ", time2 - time1)

ppv = sortperm(loadofdistrict, rev=true) #The first the most loaded matrices index
least_index_cv = ppv[M]
most_index_cv = ppv[1]
most_value_cv = loadofdistrict[most_index_cv]
least_value_cv = loadofdistrict[least_index_cv]
diff = most_value_cv - least_value_cv
#=valeur_fin=minimum(tracking)
iteration_graph=collect(1:length(tracking))
tre=tracking[1:end]
graph=plot(x=iteration_graph, y=tre, Geom.point, Theme(default_color=colorant"#E3E1E1", background_color = colorant"#A6A6A6"), Guide.xlabel("Iteration"), Guide.ylabel("Optimal Value"))
draw(SVG("graph_diff_value_para_.svg", 15cm, 20cm), graph)
=#

#For printing the results in the excel file
outfile = open("results_SA.txt", "w")
println(outfile,"File = 4")
println(outfile,"Initial cost = $Objectiveinitial ")
println(outfile,"lower bound = $lb ")
println(outfile,"Best solution found=$diff")
println(outfile,"Best solution found=$matrices")
println(outfile,"Load=$loadofdistrict")
close(outfile)
in_plateau
                end
            end
        end

        nb_plateau_respect_condition=0
        for i in 1:K2
            if acceptedmove_for_K2_plateau[i]<((epsilon/100)*L)
                nb_plateau_respect_condition=nb_plateau_respect_condition+1
            end
        end
        # STOP DECISION

        if (nb_plateau_without_improv >= K2) && (nb_plateau_respect_condition==K2)
            Terminate=true

        else
            # Proceed to the next plateau
            iteration_within_plateau=1 # reinitialize the iteration count as we ended the plateau
            acceptedmove_within_plateau = 0 # reset the number of moves on the plateau as we ended it
            T=T*alpha # decrease the temperature by alpha

            #println("T=", T)
            improv_within_plateau=false # reinitialize the fact that we have made an improvement within the plateau
            plateau_nb=plateau_nb+1
        end
    end
end

time2 = time() # END TIME
FINALTIME=time2-time1
println("Execution time: ", time2 - time1)

ppv = sortperm(loadofdistrict, rev=true) #The first the most loaded matrices index
least_index_cv = ppv[M]
most_index_cv = ppv[1]
most_value_cv = loadofdistrict[most_index_cv]
least_value_cv = loadofdistrict[least_index_cv]
diff = most_value_cv - least_value_cv
#=valeur_fin=minimum(tracking)
iteration_graph=collect(1:length(tracking))
tre=tracking[1:end]
graph=plot(x=iteration_graph, y=tre, Geom.point, Theme(default_color=colorant"#E3E1E1", background_color = colorant"#A6A6A6"), Guide.xlabel("Iteration"), Guide.ylabel("Optimal Value"))
draw(SVG("graph_diff_value_para_.svg", 15cm, 20cm), graph)
=#

#For printing the results in the excel file
outfile = open("results_SA.txt", "w