using DataFrames
using XLSX
using JuMP  # Julia for Mathematical Programming
using Gadfly
using Statistics

# Data Import
file = ("data_final.xlsx")
contiguity = DataFrame(XLSX.readtable(file, "contiguity")...)
area_bu =DataFrame(XLSX.readtable(file, "area")...)
distance_bu = DataFrame(XLSX.readtable(file, "network_distance")...)
city_subunit1 =  DataFrame(XLSX.readtable(file, "city")...)
care_load=DataFrame(XLSX.readtable(file, "workload")...)
compa = DataFrame(XLSX.readtable("agDATA.xlsx", "compatibility")...)
e2 = compa[1:84, 2:85]
#Parameters
N = 84
M = 4 # enter the number of districts to design here
s = 6/5 # enter the average speed here
e = contiguity[1:84, 2:85]
area=area_bu[1:84,2]
d=distance_bu[1:84,2:85] #d[i,j] be the network distance between subunit i and subunit j.
city_subunit = city_subunit1[1:84, 2]
number =care_load[1:84,2] #Number of patient
wt =care_load[1:84,3] # total load  OF SUBREGION i in minutes

######################################################################################
##################PHASE 1 : PAP => Incompability   ##################
######################################################################################

#1)We define Incompability
time1initial=time()

w1=0.1 #Distance weight
w2=2#Contiguity weight
w3=1#City weight

function S(i,j)
    if city_subunit[i]==city_subunit[j]
        city=1
    else
        city=0
    end
    return (w2*(1-e[i,j]))+(w3*(1-city))+(w1*d[i,j])
end

matrixIncompa = Vector{Float64}()

for i in 1:N
    ind = sum(j ≠ i ? S(i, j) : 0 for j in 1:N)
    push!(matrixIncompa, ind)
end
 #For each basic unit we compute the total number of cities with wich the given city is contiguous or in the same city

#Gives incompability score to each district

MatrixRelation=zeros(N, N)

for i in 1:N
    for j in 1:N
        if i != j
            MatrixRelation[i, j] = S(i, j)
        end
    end
end

limit=4 #The limit #Means that if not contiguous and not same city , it is in max 10km

MatrixI=zeros(N, N) #If 1 it means that incompatible
for i in 1:N
    for j in 1:N
        if MatrixRelation[i, j]>4
            MatrixI[i, j] = 1
        end
    end
end

scoreIncompa = Vector{Float64}()
for i in 1:N
    ind = sum(j ≠ i ? MatrixI[i,j] : 0 for j in 1:N)
    push!(scoreIncompa, ind)
end


matrices = Vector{Vector{Int}}(undef, M) #We create a matrix with M vectors in wich each vecotr represent a district
for i in 1:M
    matrices[i] = []
end


sorted_incompa = sortperm(scoreIncompa, rev=true)  # The first = most incompatible
i = 1
t = 1
while i <= M
    sub1 = sorted_incompa[t]  # Index of the max incompatibility
    if i == 1  # We select the most incompatible and place it in the first district / first element
        push!(matrices[i], sub1)
        t += 1
        i += 1
    else  # For the other ones, we make sure there is no contiguity with previously selected basic units
        noncontig = 0
        x = 1
        a = i - 1
        while x <= a
            if d[matrices[x][1], sub1] >= 22  # Check contiguity condition (e.g., d[i, j] >= 20)
                noncontig += 1
                if noncontig == a && x == a  # We have verified for all previous basic units
                    push!(matrices[i], sub1)
                    t += 1
                    i += 1  # Move to the next matrix
                    break
                else
                    x += 1
                end
            else  # Case where the subunit is contiguous or in the same city as a previous one
                t += 1
                break  # Exit the inner while loop
            end
        end
    end
end




######################################################################################
###############PHASE 2: Workload balancing heuristic/Initial solution#################
######################################################################################


# Initialize an empty array to hold the workloads of each basic unit
workloadofbu = []

# Loop over each basic unit and compute the total workload
for i in 1:length(number)
    push!(workloadofbu,  wt[i])
end


# Sort basic units in decreasing order of workloads
index_decreasing = sortperm(workloadofbu, rev=true) #Donne l'indice des load dans l'ordre décroissant

# Remove any existing basic units in districts from index_decreasing
for j in 1:M
    deleteat!(index_decreasing, findfirst(x -> x == matrices[j][1], index_decreasing))
end

#3) Select the first basic unit of the list and assign it to the least loaded district containing one basic unit contiguous with the first element of the list.
# Initialize workload of each district to 0
loadofdistrict = zeros(M)

# Compute the workload of each initial district
for i in 1:M
    loadofdistrict[i] = workloadofbu[matrices[i][1]]
end

a = 1
while !isempty(index_decreasing)
    i = 1
    assigned = false  # Initialize assigned flag

    while i <= M
        # Find the least loaded district
        sorted_district = sortperm(loadofdistrict, rev=false)
        least_index = sorted_district[i]

        # Check conditions for assignment
        found_assignment = false
        for j in 1:size(matrices[least_index])[1]

            #if MatrixI[index_decreasing[a], matrices[least_index][j]]==0
            if e[index_decreasing[a], matrices[least_index][j]]==1


                # Assign the first element of the list to the least loaded district
                adid=index_decreasing[a]
                append!(matrices[least_index],adid)
                deleteat!(index_decreasing, a)
                # Recompute the workload of the least loaded district
                sum_number = sum(number[matrices[least_index]])
                sum_area = sum(area[matrices[least_index]])
                cload = sum(wt[matrices[least_index]])
                loadofdistrict[least_index] = cload


                assigned = true  # Set assigned flag to true
                found_assignment = true
                break
            end
        end

        if found_assignment
            a = 1
            break
        else
            i += 1
        end
    end

    if !assigned
        a += 1
    end
end
time2initial=time()
timeinitial=time2initial-time1initial
Objectiveinitial = maximum(loadofdistrict)-minimum(loadofdistrict)
initial=copy(matrices)
allunique(matrices)
initial_load=copy(loadofdistrict)

######################################################################################
############Phase 3 : Sol Improvement Heuristic##############
######################################################################################

time1descent1=time()
 test = 0
 value1=[]
 mat_test = []
 mat_select = []
 initial_load_bu = copy(workloadofbu)
 iter = 1
 improved = 0
 xx = M
 f = 1
 pp = sortperm(loadofdistrict, rev=true)
 selected_last_min = 0
 selected_last_max =0
 limit=1000
 debut=0
 tracking=[]
 num=0
 while iter<limit
     xx = M
     f = 1
     pp = sortperm(loadofdistrict, rev=true)

     least_index = pp[xx]
     most_index = pp[f]
     if length(matrices[most_index]) == 1
         for i in 1:M-2
             if length(matrices[most_index+i])>1
                 f=f+1
                 break
             end
         end
     end
     most_index = pp[f]
     most_value = loadofdistrict[most_index]
     least_value = loadofdistrict[least_index]
     difference = most_value - least_value
     aa = copy(matrices[most_index])
     idx = argmin(abs.(workloadofbu[aa] .- difference))
     debut=debut+1

     con=true
     while con==true
         last = length(mat_select)
         if last > 1
             if mat_select[last] == aa[idx] && length(aa) > 1  ||  mat_select[last-1] == aa[idx] && length(aa) > 1
                 deleteat!(aa, idx)
                 idx = argmin(abs.(workloadofbu[aa] .- difference))

             elseif length(aa) <= 1 && xx > 2
                 xx = xx - 1
                 least_index = pp[xx]
                 most_index = pp[f]
                 least_value = loadofdistrict[least_index]
                 most_value = loadofdistrict[most_index]
                 difference = most_value - least_value
                 aa = copy(matrices[most_index])
                 idx = argmin(abs.(workloadofbu[aa] .- difference))
             elseif xx==2 && length(aa) <= 1
                 if f<M
                     f=f+1
                     most_index = pp[f]
                     least_value=pp[xx]
                     most_value = loadofdistrict[most_index]
                     difference = most_value - least_value
                     aa = copy(matrices[most_index])
                     idx = argmin(abs.(workloadofbu[aa] .- difference))

                 else
                 end

             end
         end

         selected_last_min = least_index
         selected_last_max = most_index

         jcomp = 0
         size_matrices_selected = size(matrices[least_index], 1)

         for j in 1:size_matrices_selected
             if e[aa[idx], matrices[least_index][j]] == 0
                 jcomp += 1
                 break
             end
         end

         if jcomp > 0
             init_mostload=loadofdistrict[most_index]
             init_leastload=loadofdistrict[least_index]
             if length(value1)==0
                 difference_test = maximum(loadofdistrict)-minimum(loadofdistrict)
             else
                 difference_test=value1[length(value1)]
             end
             push!(matrices[least_index], aa[idx])
             sup = findfirst(matrices[most_index] .== aa[idx])
             deleteat!(matrices[most_index], sup)

             cload_least = sum(wt[matrices[least_index]])
             cload_most = sum(wt[matrices[most_index]])
             ll1=cload_least
             ll2 =cload_most
             loadofdistrict[least_index] = ll1
             loadofdistrict[most_index] = ll2
             objective=maximum(loadofdistrict)-minimum(loadofdistrict)
             #Nouvelle valeur
             num=num+1
             append!(tracking,objective)
             if difference_test>=objective
                 # Accept the move
                 push!(value1,objective)
                 improved=improved+1
                 push!(mat_select, aa[idx])

                 xx=M
                 f=1
                 con=false
                 break
             else
                 # Revert the move
                 push!(matrices[most_index], aa[idx])
                 sup = findfirst(matrices[least_index] .== aa[idx])
                 deleteat!(matrices[least_index], sup)
                 loadofdistrict[least_index] =init_leastload
                 loadofdistrict[most_index]=init_mostload
                 push!(mat_select, aa[idx])
                 #On doit regarder si on peut changer un autre élément  #peut etre faire matrice
                 global iter=iter+1
                 if iter >= limit
                      # Exit the loop when iter reaches or exceeds 100
                 break
                 end
                 if length(aa) > 1 #Ici on tranfere entre mm matrice donc mm diff
                     deleteat!(aa, idx)
                     idx = argmin(abs.(workloadofbu[aa] .- difference))
                     last = length(mat_select)
                     break
                 end
             end
         elseif length(aa) > 1 #Ici on tranfere entre mm matrice donc mm diff

             deleteat!(aa, idx)
             idx = argmin(abs.(workloadofbu[aa] .- difference))
             last = length(mat_select)
             break

         elseif length(aa) <= 1
             if xx > 2 #least index change #diff change
                 #la diff est mm , on va mettre vers la deuxieme moins loadé
                 aa = copy(matrices[most_index])
                 pp = sortperm(loadofdistrict, rev=true)
                 xx -= 1
                 least_index = pp[xx]
                 difference = pp[f] - least_index
                 idx = argmin(abs.(workloadofbu[aa] .- difference))
             else
                 break
             end
         end
     end
 end


 time2descent1=time()
 timedescent=time2descent1-time1descent1
 ppv = sortperm(loadofdistrict, rev=true) #The first the most loaded matrices index
 least_index_cv = ppv[M]
 most_index_cv = ppv[1]
 most_value_cv = loadofdistrict[most_index_cv]
 least_value_cv = loadofdistrict[least_index_cv]
ObjectiveDescent = most_value_cv - least_value_cv

#####Compute the travel load

function travelload(N,A)
    # Calculate travel load based on number of people and area
    return (0.75 * sqrt(N*A) *s)
end


#FIRST DEFINITION

firstdefinition_load_matrices=[]
for i in 1:M
    sum_number = sum(number[matrices[i]])
    sum_area = sum(area[matrices[i]])
    sum_cload = sum(wt[matrices[i]])
    ll1= (travelload(sum_area, sum_number) + sum_cload)
    push!(firstdefinition_load_matrices, ll1)
end

objective_first=maximum(firstdefinition_load_matrices)-minimum(firstdefinition_load_matrices)



#SECOND DEFINITION
workloadofbu = []

# Loop over each basic unit and compute the total workload #We change the first load matrices
for i in 1:length(number)
    tot_load = travelload(number[i], area[i]) + wt[i]
    push!(workloadofbu, tot_load)
end

seconddefinition_load_matrices = []
isize=[]
for i in 1:M
    global distance = 0
    global selection1=1 #indices in selected_nod
    selected_nod = copy(matrices[i])
    if size(matrices[i])[1]==1
        push!(seconddefinition_load_matrices, sum(workloadofbu[matrices[i]]))
        workloadofbu[30]
    else
        while !isempty(selected_nod)

            selection1v=findfirst(x -> x == selected_nod[selection1], matrices[i])
            global selection1value=selected_nod[selection1]
            deleteat!(selected_nod, selection1)
            if isempty(selected_nod)
                last_selec=copy(matrices[i])
                selection1_l=findfirst(x -> x == selection1value, matrices[i])
                deleteat!(last_selec, selection1_l)
                selection2 = argmin(d[last_selec, selection1value])[1] #Indices in selected_nod
                selected2value=last_selec[selection2]
                value = d[selected2value,selection1value]
                global distance += value
                break
            else
                selection2 = argmin(d[selected_nod, selection1value])[1]#Indices in selected_nod
                selected2value=selected_nod[selection2]
                value = d[selected2value,selection1value]
                global distance += value
                selection1=selection2
                test=test+1
            end
        end
        push!(seconddefinition_load_matrices, sum(workloadofbu[matrices[i]]) + distance)
        push!(isize,size(isize)[1])
    end
end

objective_second=maximum(seconddefinition_load_matrices)-minimum(seconddefinition_load_matrices)


#########Print Results
outfile = open("results_Comparison.txt", "w")
println(outfile,"File = 4")
println(outfile,"InitialObjective = $Objectiveinitial ")
println(outfile,"DescentObjective = $ObjectiveDescent ")
println(outfile,"ObjectiveFirstMethod = $objective_first ")
println(outfile,"ObjectiveSecondMethod = $objective_second ")
println(outfile,"Best solution found=$matrices")
println(outfile,"LoadOzturk=$loadofdistrict")
println(outfile,"LoadFirst=$firstdefinition_load_matrices")
println(outfile,"LoadSecond=$seconddefinition_load_matrices")
close(outfile)
[i])[1]==1
        push!(seconddefinition_load_matrices, sum(workloadofbu[matrices[i]]))
        workloadofbu[30]
    else
        while !isempty(selected_nod)

            selection1v=findfirst(x -> x == selected_nod[selection1], matrices[i])
            global selection1value=selected_nod[selection1]
            deleteat!(selected_nod, selection1)
            if isempty(selected_nod)
                last_selec=copy(matrices[i])
                selection1_l=findfirst(x -> x == selection1value, matrices[i])
           